"""Various constants"""

DEVICE_URL = "https://ads-field.aylanetworks.com"
LOGIN_URL = "https://user-field.aylanetworks.com"
SHARK_APP_ID = "Shark-Android-field-id"
SHARK_APP_SECRET = "Shark-Android-field-Wv43MbdXRM297HUHotqe6lU1n-w"

